#!/usr/bin/env python

import euclid
import math
from constants import *

class LT(object):
    """
    Coordinates are like a screen, ie, x increases from left to right and y
    increases from top to bottom.
    """
    def __init__(self):
        self.results = []
        self.drawing = False
        self.home()

    def normalize(self, heading):
        while heading > math.pi:
            heading -= 2 * math.pi
        while heading <= -math.pi:
            heading += 2 * math.pi
        return heading
    
    def recalculateHeading(self):
        self.cosh = math.cos(self.heading)
        self.sinh = math.sin(self.heading)

    def record(self):       
        if not self.drawing: return
        w = GRANULARITY * 0.4
        lx = self.x + w * self.sinh
        ly = self.y + w * -self.cosh
        rx = self.x + w * -self.sinh
        ry = self.y + w * self.cosh

        r = self.results
        if len(r) < 2:
            self.mx, self.my = self.x, self.y
            r.append((lx, ly, rx, ry))
            return

        # get the end points of the last left and right lines
        lx1, ly1, rx1, ry1 = r[-2]
        lx2, ly2, rx2, ry2 = r[-1]

        # get the left and right coordinates of the start point of the new line
        lx3 = self.mx + w * self.sinh
        ly3 = self.my + w * -self.cosh
        rx3 = self.mx + w * -self.sinh
        ry3 = self.my + w * self.cosh

        # find where the left lines intersect
        lp1 = euclid.Point2(lx1, ly1)
        lp2 = euclid.Point2(lx2, ly2)
        lp3 = euclid.Point2(lx3, ly3)
        lp4 = euclid.Point2(lx, ly)
        l1 = euclid.Line2(lp1, lp2)
        l2 = euclid.Line2(lp3, lp4)
        lp = l1.intersect(l2)
        if not lp:
            lp = lp2

        # find where the right lines intersect
        rp1 = euclid.Point2(rx1, ry1)
        rp2 = euclid.Point2(rx2, ry2)
        rp3 = euclid.Point2(rx3, ry3)
        rp4 = euclid.Point2(rx, ry)
        r1 = euclid.Line2(rp1, rp2)
        r2 = euclid.Line2(rp3, rp4)
        rp = r1.intersect(r2)
        if not rp:
            rp = rp2

        # replace the last points with the points of intersection
        r[-1] = (lp.x, lp.y, rp.x, rp.y)

        # add the new points
        self.mx, self.my = self.x, self.y
        r.append((lx, ly, rx, ry))
        
    def home(self):
        self.x = 0.0
        self.y = 0.0
        self.heading = 0.0
        self.recalculateHeading()
        self.record()
                
    def forward(self, dist):
        self.x += dist * self.cosh
        self.y += dist * self.sinh
        self.record()

    def back(self, dist):
        self.forward(-dist)
        self.record()

    def right(self, angle):
        self.heading = self.normalize(self.heading + angle)
        self.recalculateHeading()
##        self.record()

    def left(self, angle):
        self.right(-angle)
##        self.record()

    def face(self, angle):
        self.heading = self.normalize(angle)
        self.recalculateHeading()
##        self.record()

    def goto(self, x, y):
        self.x = x
        self.y = y
        self.record()
        
if __name__ == '__main__':
    lt = LT()
    lt.drawing = True
    lt.home()
    lt.forward(50)
    lt.left(math.pi/2)
    lt.forward(50)
