import euclid
import pygame
import os
import sys
from level_turtle import LT
from OpenGL.GL import *
from OpenGL.GLU import *
from constants import *
import math


# --- begin section ----------------------------------------------------------
# Code from http://www.pygame.org/wiki/SimpleOpenGL2dClasses?parent=CookBook

def initializeDisplay(w, h):
    pygame.display.set_mode((w,h), pygame.OPENGL|pygame.DOUBLEBUF)
 
    glClearColor(0.0, 0.0, 0.0, 1.0)
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT)
 
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity();
    # this puts us in quadrant 1, rather than quadrant 4
    gluOrtho2D(0, w, h, 0)
    glMatrixMode(GL_MODELVIEW)

    glDisable(GL_CULL_FACE)
    
    # set up texturing
    glEnable(GL_TEXTURE_2D)
    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

 
def render_init(w,h):
    """Finds the smallest available resolution that fits the desired
    viewfield."""
    pygame.init()
    modelist = pygame.display.list_modes()
    nextmode = []
    for l in modelist:
        if l[0]>=w and l[1]>=h:
            nextmode.append(l)
    bestx, besty = -1,-1
    for l in nextmode:
        if bestx==-1 or bestx>=l[0]:
            if besty==-1 or besty>=l[1]:
                bestx, besty = l[0],l[1]
 
    initializeDisplay(bestx, besty)

 
def loadImage(image):
    textureSurface = pygame.image.load(image)
 
    textureData = pygame.image.tostring(textureSurface, "RGBA", 1)
 
    width = textureSurface.get_width()
    height = textureSurface.get_height()
 
    texture = glGenTextures(1)
    glBindTexture(GL_TEXTURE_2D, texture)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA,
        GL_UNSIGNED_BYTE, textureData)
 
    return texture, width, height

 
def drawImage(x, y, texture, width, height, theta):
    glEnable(GL_BLEND)
    glDisable(GL_ALPHA_TEST)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    
    theta = math.degrees(theta)
    glPushMatrix()
    glColor(1.0, 1.0, 1.0)
    glBindTexture(GL_TEXTURE_2D, texture)
    glTranslate(x, y, 0)
    glRotate(theta, 0, 0, -1)
    glTranslate(-width/2, -height/2, 0)
    glBegin(GL_QUADS)
 
    # Bottom Left Of The Texture and Quad
    glTexCoord2f(0, 0); glVertex2f(0, 0)
 
    # Top Left Of The Texture and Quad
    glTexCoord2f(0, 1); glVertex2f(0, height)
 
    # Top Right Of The Texture and Quad
    glTexCoord2f(1, 1); glVertex2f( width,  height)
 
    # Bottom Right Of The Texture and Quad
    glTexCoord2f(1, 0); glVertex2f(width, 0)
    glEnd()
    glPopMatrix()

    glDisable(GL_BLEND)

# --- end section ------------------------------------------------------------

def toGrid(x, y):
    """Converts to grid coordinates."""
    if x < 0:
        x -= GRANULARITY
    x = int(x / GRANULARITY)
    if y < 0:
        y -= GRANULARITY
    y = int(y / GRANULARITY)
    return (x, y)

        
def boundingBox(pts):
    """Returns the bounding box of a sequence of points."""
    maxx = maxy = -1e10
    minx = miny = 1e10
    for x, y in pts:
        maxx = max(maxx, x)
        minx = min(minx, x)
        maxy = max(maxy, y)
        miny = min(miny, y)
    cx = (maxx + minx) / 2.0
    cy = (maxy + miny) / 2.0
    return ((minx, miny), (maxx, maxy), (cx, cy))


def drawGrid(x, y):
    """Draws the background "parallax" grid."""
    glDisable(GL_TEXTURE_2D)
    glEnable(GL_BLEND)
    STEP = min(WIDTH / 6, HEIGHT / 6)
    x = STEP - (x % STEP)
    y = STEP - (y % STEP)
    glColor(0.125, 0.125, 0.25, 1.0)
    glBegin(GL_LINES)
    ix = x
    while ix < x + WIDTH:
        glVertex(ix, 0)
        glVertex(ix, HEIGHT-1)
        ix += STEP
    iy = y
    while iy < y + HEIGHT:
        glVertex(0, iy)
        glVertex(WIDTH-1, iy)
        iy += STEP
    glEnd()
    glEnable(GL_TEXTURE_2D)


def drawArrow():        
    """Draws the directional arrow."""
    glDisable(GL_TEXTURE_2D)
    glEnable(GL_BLEND)
    x = Mouse.pos[0] - WIDTH/2
    y = Mouse.pos[1] - HEIGHT/2
    theta = math.atan2(y, x)
    d = math.sqrt(x*x + y*y)
    d = min(100, d)
    pygame.mouse.set_pos((WIDTH/2 + d*math.cos(theta), HEIGHT/2 + d*math.sin(theta)))
    glPushMatrix()
    glTranslatef(WIDTH/2, HEIGHT/2, 0)
    glRotatef(math.degrees(theta), 0, 0, 1)
    glColor(0.5, 1.0, 0.0, 0.333)
    glBegin(GL_QUADS)
    glVertex(-10, -10)
    glVertex(d, -10)
    glVertex(d, 10)
    glVertex(-10, 10)
    glEnd()
    n = d + max(10, d*0.25)
    glBegin(GL_POLYGON)
    glVertex(d, -10)
    glVertex(d, -20)
    glVertex(n, 0)
    glVertex(d, 20)
    glVertex(d, 10)
    glEnd()
    glPopMatrix()

        
class Mouse:
    pos = (0.0, 0.0)


class Player(object):

    def __init__(self, x, y):
        self.image = loadImage('wand.bmp')
        self.x = float(x)
        self.y = float(y)
        self.texture, self.width, self.height = self.image
        self.rotate = 0.0
        self.spinSpeed = math.pi/180.0

    def update(self):
        self.rotate += self.spinSpeed
        while self.rotate > math.pi:
            self.rotate -= 2 * math.pi
        while self.rotate <= math.pi:
            self.rotate += 2 * math.pi

        x, y = Mouse.pos
        x -= WIDTH / 2.0
        y -= HEIGHT / 2.0
        angle = math.atan2(y, x)
        length = (math.sqrt(x*x + y*y) * 0.125) - 1
        length = max(0, min(4, length))
        dx = length * math.cos(angle)
        dy = length * math.sin(angle)
        self.x += dx
        self.y += dy
        
    def draw(self):
        # the player is always drawn in the middle of the screen
        drawImage(WIDTH/2.0, HEIGHT/2.0, self.texture, self.width, self.height, self.rotate)
        

class Tile(object):

    def __init__(self):
        self.paths = []

    def addPath(self):
        self.paths.append([])

    def addQuad(self, lt, rt):
        self.paths[-1].append((lt, rt))
                
    def checkCollision(self, path, geom):
        hits = []
        lines = []
        x1, y1 = geom[0]
        for x, y in geom[1:]:
            lines.append(euclid.LineSegment2(euclid.Point2(x1, y1), euclid.Point2(x, y)))
            x1, y1 = x, y   
        for line in lines:
            for l, r in path:
                pt = l.intersect(line)
                if pt:
                    hits.append(pt)
                pt = r.intersect(line)
                if pt:
                    hits.append(pt)
        return hits

    def checkCollisions(self, geom):
        hits = []
        for path in self.paths:
            pts = self.checkCollision(path, geom)
            if pts:
                hits.append((path, pts))
        return hits                        
            
    def drawBorder(self, path):
        glBegin(GL_LINES)
        for l, r in path:
            glVertex(l.p1.x, l.p1.y)
            glVertex(l.p2.x, l.p2.y)
            glVertex(r.p1.x, r.p1.y)
            glVertex(r.p2.x, r.p2.y)
        glEnd()
        
    def drawBorders(self):
        for path in self.paths:
            self.drawBorder(path)

    def drawPath(self, path):
        glBegin(GL_QUADS)
        for l, r in path:
            glVertex(l.p1.x, l.p1.y)
            glVertex(l.p2.x, l.p2.y)
            glVertex(r.p2.x, r.p2.y)
            glVertex(r.p1.x, r.p1.y)
        glEnd()
        
    def drawPaths(self):
        for path in self.paths:
            self.drawPath(path)


class Grid(object):

    def __init__(self):
        self.grid = {}
        
    def addPath(self, path):
        lx1, ly1, rx1, ry1 = path[0]
        for i in range(1, len(path)):
            lx, ly, rx, ry = path[i]
            aabb = boundingBox([(lx, ly), (rx, ry), (lx1, ly1), (rx1, ry1)])
            cx, cy = aabb[-1]
            cx, cy = toGrid(cx, cy)
            if not (cx, cy) in self.grid:
                tile = Tile()
                tile.addPath()
                self.grid[(cx, cy)] = tile
            else:
                tile = self.grid[(cx, cy)]
            lt = euclid.LineSegment2(euclid.Point2(lx1, ly1), euclid.Point2(lx, ly))
            rt = euclid.LineSegment2(euclid.Point2(rx1, ry1), euclid.Point2(rx, ry))
            tile.addQuad(lt, rt)
            lx1, ly1, rx1, ry1 = lx, ly, rx, ry

    def checkCollisions(self, geom):
        # get all the points that are in collision
        aabb = boundingBox(geom)
        cx, cy = aabb[-1]
        cx, cy = toGrid(cx, cy)
        hits = []
        for x in range(cx-1, cx+2):
            for y in range(cy-1, cy+2):
                if (x, y) in self.grid:
                    tile = self.grid[(x, y)]
                    hits.extend(tile.checkCollisions(geom))
        return hits
    
    def drawBorders(self):
        glDisable(GL_BLEND)
        glDisable(GL_ALPHA_TEST)
        glDisable(GL_TEXTURE_2D)
        glEnable(GL_LINE_SMOOTH)
        glLineWidth(3.0)
        glColor(1.0, 1.0, 0.0)
        for k, tile in self.grid.iteritems():
            x, y = k
            x = (x * GRANULARITY)
            y = (y * GRANULARITY)
            if x + 2 * GRANULARITY < self.worldx: continue
            if x - GRANULARITY > self.worldx + WIDTH: continue
            if y + 2 * GRANULARITY < self.worldy: continue
            if y - GRANULARITY > self.worldy + HEIGHT: continue
            tile.drawBorders()
        glLineWidth(1.0)
        glEnable(GL_TEXTURE_2D)

    def drawPaths(self):
        glDisable(GL_BLEND)
        glDisable(GL_ALPHA_TEST)
        glDisable(GL_TEXTURE_2D)
        glEnable(GL_LINE_SMOOTH)
        glColor(0.15, 0.15, 0.15)
        for k, tile in self.grid.iteritems():
            x, y = k
            x = (x * GRANULARITY)
            y = (y * GRANULARITY)
            if x + 2 * GRANULARITY < self.worldx: continue
            if x - GRANULARITY > self.worldx + WIDTH: continue
            if y + 2 * GRANULARITY < self.worldy: continue
            if y - GRANULARITY > self.worldy + HEIGHT: continue
            tile.drawPaths()
        glDisable(GL_LINE_SMOOTH)
        glEnable(GL_TEXTURE_2D)
            
    def drawGrid(self):
        glDisable(GL_TEXTURE_2D)
        glEnable(GL_LINE_SMOOTH)
        glColor(0.0, 0.5, 0.0)
        for k in self.grid.iterkeys():
            x, y = k
            x = (x * GRANULARITY)
            y = (y * GRANULARITY)
            if x + 2 * GRANULARITY < grid.worldx: continue
            if x - GRANULARITY > grid.worldx + WIDTH: continue
            if y + 2 * GRANULARITY < grid.worldy: continue
            if y - GRANULARITY > grid.worldy + HEIGHT: continue
            glBegin(GL_LINE_LOOP)
            glVertex(x, y)
            glVertex(x + GRANULARITY, y)
            glVertex(x + GRANULARITY, y + GRANULARITY)
            glVertex(x, y + GRANULARITY)
            glEnd()
        glDisable(GL_LINE_SMOOTH)
        glEnable(GL_TEXTURE_2D)

    def draw(self, x, y):
        """Draws the entire (visible) grid."""
        glPushMatrix()
        glTranslatef(-x, -y, 0.0)
        
        # draw the path's borders
        self.drawBorders()

        # draw the path itself
        self.drawPaths()

        # draw the sparse grid which contains the paths
        if debug:
            self.drawGrid()
        glPopMatrix()
        

def createLevel():
    """Draws the level using 'turtle' graphics.
    
    Ideally there would be multiple paths in the level, but that's not an
    option until:

    a) I load the levels externally
    b) I create a level by hand
    c) I figure out a sensible way of checking if a collision point is
       obscured by a quad that doesn't belong to its path.
    """
    lt = LT()
    lt.goto(0, 0)
    lt.drawing = True
    lt.forward(0)
    for j in range(4):
        for i in range(4):
            lt.forward(150)
        for i in range(18):
            lt.forward(20)
            lt.left(math.pi/72)
        for i in range(18):
            lt.forward(20)
            lt.right(math.pi/72)
        for i in range(18):
            lt.forward(20)
            lt.right(math.pi/72)
        for i in range(18):
            lt.forward(20)
            lt.left(math.pi/72)
        for i in range(4):
            lt.forward(150)
        for i in range(18):
            lt.forward(40)
            lt.right(math.pi/36)
    lt.forward(10)        
    grid = Grid()
    grid.addPath(lt.results)
    return grid


def main():
    render_init(WIDTH, HEIGHT)
    clock = pygame.time.Clock()
    pygame.mouse.set_visible(False)
    grid = createLevel()    
    player = Player(0.0, 0.0)
    pygame.mouse.set_pos(WIDTH/2.0, HEIGHT/2.0)
    
    while True:
        clock.tick(HERTZ)
        for event in pygame.event.get():
            # quit if the user closed the window
            if event.type == pygame.QUIT:
                pygame.quit()
                return
            # quit if the user pressed escape
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    pygame.quit()
                    return
            # figure out where the mouse is
            elif event.type == pygame.MOUSEMOTION:
                Mouse.pos = event.pos

        # move the player
        oldx, oldy = player.x, player.y
        player.update()

        # collision detection (a simple line ... bet you didn't expect that)
        dx = math.cos(player.rotate) * 64
        dy = math.sin(player.rotate) * 64
        geom = [(player.x - dx, player.y + dy), (player.x + dx, player.y - dy)]
        collisions = grid.checkCollisions(geom)

        # severely simplified collision response: rotate the other way and
        # bounce back the way we came at half the speed
        if collisions:
            player.spinSpeed = -player.spinSpeed
            player.rotate += player.spinSpeed
            player.x, player.y = oldx, oldy
            x, y = Mouse.pos
            x -= WIDTH/2
            y -= HEIGHT/2
            x *= -0.5
            y *= -0.5
            newpos = (WIDTH/2+x, HEIGHT/2+y)
            pygame.mouse.set_pos(*newpos)
            Mouse.pos = newpos

        # update the world to match the player's movement
        x, y = player.x, player.y
        x -= WIDTH/2
        y -= HEIGHT/2
        grid.worldx, grid.worldy = x, y
        
        glClearColor(0.25, 0.25, 0.50, 1.0)
        glClear(GL_COLOR_BUFFER_BIT)

        # draw the background "parallax" grid, then the level itself
        drawGrid(x * 0.5, y * 0.5)
        grid.draw(x, y)
        
        # draw the player on top of it all
        glEnable(GL_TEXTURE_2D)
        player.draw()

        # draw the directional arrow
        drawArrow()

        pygame.display.flip()
    
 
if __name__=="__main__":
    try:
        import psyco
        psyco.full()
    except ImportError:
        pass
    debug = '-d' in sys.argv
    main()
